// Factory implementation, should be included in kernel module.

#include <iostream>
#include <cstdio>

namespace regime {

  using std::string;

  template <class KEY, class BASE>
    class factory_impl : public factory<KEY,BASE> {
    public:

      //==================================================
      // CLASS: factory
      //==================================================
      factory_impl() {
        m_libraries[""] = keys();
      }
    
      ~factory_impl() {
        for (libinfos::iterator i = m_library_infos.begin(); i != m_library_infos.end(); ++i) {
          assert((*i).second.is_auto == true);
        }
      }

      //==================================================
      // Internal data
      //==================================================
    private:

      struct library_info {
        library_info()
          : handle(0), is_auto(false)
          {}
        library_info(void* _handle, bool _is_auto)
          : handle(_handle), is_auto(_is_auto)
          {}
        void* handle;
        bool is_auto;
      };
      typedef std::map<std::string, library_info> libinfos;

      creators m_creators;
      keys m_keys;
      libraries m_libraries;
      std::string m_current_library;
      libinfos m_library_infos;
      load_strategy<KEY>* m_load_strategy;

    protected:
    
      //==================================================
      // Implementation of internal methods
      //==================================================

      //--------------------------------------------------
      // METHOD: internal_lookup_creator
      // DESCRIPTION:
      //--------------------------------------------------
      creators::const_iterator internal_lookup_creator(const KEY& key) {

        creators::const_iterator i = m_creators.find(key);
        if (i == m_creators.end()) {
          if (m_load_strategy == 0)
            return i;
          string library = m_load_strategy->find(key);
          if (library == "")
            return i;
          internal_load_library(library, true);
          i = m_creators.find(key);
          assert(i != m_creators.end());
        }
        return i;
      }
      
      //--------------------------------------------------
      // METHOD: internal_create
      // DESCRIPTION:
      //--------------------------------------------------
      BASE* internal_create(const KEY& key) {   
        creators::const_iterator i = internal_lookup_creator(key);
        return i != m_creators.end() ? (*i).second->create() : 0;
      }
      
      //--------------------------------------------------
      // METHOD: internal_create
      // DESCRIPTION:
      //--------------------------------------------------
#ifdef REGIME_FACTORY_PARAMS
      BASE* internal_create(const KEY& key, const parameters& params) { 
        creators::const_iterator i = internal_lookup_creator(key);
        return i != m_creators.end() ? (*i).second->create(params) : 0;
      }
#endif
      
      //--------------------------------------------------
      // METHOD: internal_clean
      // DESCRIPTION:
      //--------------------------------------------------
      void internal_clean() {   
        for (creators::iterator i = m_creators.begin(); i != m_creators.end(); ++i) {
          (*i).second->clean();
        }
      }
      
      //--------------------------------------------------
      // METHOD: internal_clean(key)
      // DESCRIPTION:
      //--------------------------------------------------
      void internal_clean(const KEY& key) {     
        creators::iterator i = m_creators.find(key);
        assert(i != m_creators.end());
        (*i).second->clean();
      }
      
      //--------------------------------------------------
      // METHOD: internal_load_key
      // DESCRIPTION:
      //--------------------------------------------------
      void internal_load_key(const KEY& key) {
        
        creators::const_iterator i = internal_lookup_creator(key);
        if (i != m_creators.end())
          (*i).second->load();
      }
      
      //--------------------------------------------------
      // METHOD: internal_get_current_library
      //--------------------------------------------------
      const string& internal_get_current_library() const {
        return m_current_library;
      }
      
      //--------------------------------------------------
      // METHOD: internal_load_library
      // DESCRIPTION:
      //--------------------------------------------------
      void internal_load_library(const string& library, bool is_auto) {
        
        string save_library = m_current_library;
        m_current_library = library;
        
        assert(m_libraries.find(library) == m_libraries.end());
        m_libraries[library] = keys();
        
        void* library_handle = library_loader::load(library);
        assert(m_library_infos.find(library) == m_library_infos.end());
        m_library_infos[library] = library_info(library_handle, is_auto);
        
        m_current_library = save_library;
      }
      
      //--------------------------------------------------
      // METHOD: internal_unload_library
      // DESCRIPTION:
      //--------------------------------------------------
      void internal_unload_library(const string& library) {
        
        string save_library = m_current_library;
        m_current_library = library;
        
        libinfos::iterator i = m_library_infos.find(library);
        assert(i != m_library_infos.end());
        void* library_handle = (*i).second.handle;
        library_loader::unload(library_handle);
        m_library_infos.erase(library);
        
        assert(m_libraries.find(library) != m_libraries.end());
        assert((*m_libraries.find(library)).second.size() == 0);
        m_libraries.erase(library);
        
        m_current_library = save_library;
      }
      
      //--------------------------------------------------
      // METHOD: internal_unload_key
      // DESCRIPTION:
      //--------------------------------------------------
      void internal_unload_key(const KEY& key) {
        
        creators::iterator i = m_creators.find(key);
        assert(i != m_creators.end());
        (*i).second->unload();
        const string& library = (*i).second->get_library();
        bool all_unloaded = true;
        for (i = m_creators.begin(); i != m_creators.end(); ++i) {
          if ((*i).second->is_loaded())
            all_unloaded = false;
        }
        if (all_unloaded)
          internal_unload_library(library);
      }
      
      //--------------------------------------------------
      // METHOD: internal_set_load_strategy
      //--------------------------------------------------
      load_strategy<KEY>* internal_set_load_strategy(load_strategy<KEY>* strategy) {    
        load_strategy<KEY>* save_strategy = m_load_strategy;
        m_load_strategy = strategy;
        return save_strategy;
      }
      
      //--------------------------------------------------
      // METHOD: internal_query_libraries
      //--------------------------------------------------
      const libraries& internal_query_libraries() const {
        return m_libraries;
      }
      
      //--------------------------------------------------
      // METHOD: internal_query_library
      //--------------------------------------------------
      const keys& internal_query_library(const string& library) const { 
        libraries::const_iterator i = m_libraries.find(library);
        assert(i != m_libraries.end());
        return (*i).second;
      }
      
      //--------------------------------------------------
      // METHOD: internal_query_keys
      //--------------------------------------------------
      const keys& internal_query_keys() const {
        return m_keys;
      }
      
      //--------------------------------------------------
      // METHOD: internal_register_creator
      //--------------------------------------------------
      void internal_register_creator(const KEY& key, creator* ctor) {
        
        // register the creator
        assert(m_creators.find(key) == m_creators.end());
        m_creators[key] = ctor;
        assert(m_creators.find(key) != m_creators.end());
        
        // register the key to keys
        assert(m_keys.find(key) == m_keys.end());
        m_keys.insert(key);
        
        // register the key to the library
        assert(m_libraries.find(get_current_library()) != m_libraries.end());
        keys& library_keys = m_libraries[get_current_library()];
        assert(library_keys.find(key) == library_keys.end());
        library_keys.insert(key);
      }
      
      //--------------------------------------------------
      // METHOD: internal_unregister_creator
      //--------------------------------------------------
      void internal_unregister_creator(const KEY& key, creator* ctor) {
        
        // While a library is loaded explicitly, it is unloaded
        // implicitly, and usually before the factory.
        // Thus, when a creator gets unloaded, we may not know
        // to which library it belongs.
        // In this case we make no checks.
        const string& library = ctor->get_library();
        assert(get_current_library() == "" ||
          get_current_library() == library);
        
        // unregister the key from the library
        assert(m_libraries.find(library) != m_libraries.end());
        keys& library_keys = m_libraries[library];
        assert(library_keys.find(key) != library_keys.end());
        library_keys.erase(key);
        
        // unregister the key from keys
        assert(m_keys.find(key) != m_keys.end());
        m_keys.erase(key);
        
        // unregister the creator
        assert(m_creators.find(key) != m_creators.end());
        assert((*m_creators.find(key)).second == ctor);
        m_creators.erase(key);
      }
  };
  
  //==================================================
  // Instance handling
  //==================================================
  template <class KEY, class BASE>
    factory<KEY,BASE>& factory<KEY,BASE>::get_instance() {    
    static factory<KEY,BASE>* instance;
    if (instance == 0)
      instance = new factory_impl<KEY,BASE>();
    return *instance;
  }
    
}
