#if defined(_MSC_VER) 
# if defined(_USRDLL)
#   if defined(KERNEL_EXPORTS)
#     define KAPI _declspec(dllexport)
#   else
#     define KAPI _declspec(dllimport)
#   endif
# else
#   define KAPI
# endif
# pragma warning(disable: 4786)
#else
# define KAPI
#endif

#include <map>
#include <set>
#include <string>
#include <cassert>

// Uncomment this or pass -DREGIME_FACTORY_PARAMS
// to be able to pass parameter to the factory
//#define REGIME_FACTORY_PARAMS

namespace regime {

  //==================================================
  // CLASS: creator_base
  //==================================================
  template <class BASE>
    class creator_base {
      //==================================================
      // GROUP: constructors/destructors
      //==================================================
    public:
      creator_base(const std::string& libname)
      : m_library(libname)
      , m_is_loaded(!libname.empty())
      {
      }
      virtual ~creator_base() {}

      //==================================================
      // GROUP: library handling
      //==================================================
    public:
      const std::string& get_library() const { return m_library; }
      bool is_loaded() const { return m_is_loaded; }
      void load() { m_is_loaded = true; }
      void unload() { m_is_loaded = false; }

      //==================================================
      // GROUP: product creation
      //==================================================
    public:
      virtual BASE* create() const = 0;
#ifdef REGIME_FACTORY_PARAMS
      typedef typename BASE::FactoryParams parameters;
      virtual BASE* create(const parameters& params) const = 0;
#endif
      virtual void clean() = 0;

      //==================================================
      // GROUP: internal data
      //==================================================
    private:
      const std::string m_library;
      bool m_is_loaded;
    };

  //==================================================
  // CLASS: creator
  //==================================================
  template <class KEY, class BASE, class PRODUCT>
    class creator : public creator_base<BASE> {
      //==================================================
      // GROUP: construction/destruction
      //==================================================
    public:
      creator(const KEY& key);
      ~creator();
  
      //==================================================
      // GROUP: product creation
      //==================================================
    public:
      BASE* create() const {
        return new PRODUCT;
      }
#ifdef REGIME_FACTORY_PARAMS
      BASE* create(const parameters& p) const {
        return new PRODUCT(p);
      }
#endif
      void clean() {
        // Nothing to do here
      }
      //==================================================
      // GROUP: library handling
      //==================================================
    public:
      const KEY& get_key() const {
        return m_key;
      }
    private:
        const KEY m_key;
    };

  //==================================================
  // CLASS: singleton_creator
  //==================================================
  template <class KEY, class BASE, class PRODUCT>
    class singleton_creator : public creator_base<BASE> {
      //==================================================
      // GROUP: construction/destruction
      //==================================================
    public:
      singleton_creator(const KEY& key);
      ~singleton_creator();

      //==================================================
      // GROUP: product creation
      //==================================================
    public:
      BASE* create() const {
        if (m_product == 0)
          m_product = new PRODUCT;
        return m_product;
      }
#ifdef REGIME_FACTORY_PARAMS
      BASE* create(const typename BASE::FactoryParams& params) const {
        if (m_product == 0)
          m_product = new PRODUCT(params);
        return m_product;
      }
#endif
      void clean() {
        m_product = 0;
      }
    private:
      mutable PRODUCT* m_product;
      const KEY m_key;
    };

  //==================================================
  // CLASS: load_strategy
  //==================================================
  template <class KEY>
    class load_strategy {
    public:
      virtual ~load_strategy() {}
      virtual std::string find(const KEY& key) = 0;
    };
  
  //==================================================
  // CLASS: library_loader
  //==================================================
  class KAPI library_loader {
  public:
    static void* load(const std::string& library);
    static void unload(void* library_handle);
  };

  //==================================================
  // CLASS: factory
  //==================================================
  template <class KEY, class BASE>
    class factory {
    public:
      typedef creator_base<BASE> creator;
      typedef creator_base<BASE>::parameters parameters;
      typedef std::set<KEY> keys;
      typedef std::map<std::string, keys> libraries;
      typedef std::map<KEY, creator*> creators;
    public:
      static BASE* create(const KEY& key) {
        return get_instance().internal_create(key);
      }
#ifdef REGIME_FACTORY_PARAMS
      static BASE* create(const KEY& key, const parameters& p) {
        return get_instance().internal_create(key, p);
      }
#endif

      static void clean() {
        get_instance().internal_clean();
      }
      static void clean(const KEY& key) {
        get_instance().internal_clean(key);
      }

      static const std::string& get_current_library() {
        return get_instance().internal_get_current_library();
      }
      static void load_library(const std::string& library) {
        get_instance().internal_load_library(library, false);
      }
      static void load_key(const std::string& key) {
        get_instance().internal_load_key(key);
      }
      static void unload_library(const std::string& library) {
        get_instance().internal_unload_library(library);
      }
      static void unload_key(const std::string& key) {
        get_instance().internal_unload_key(key);
      }
      static load_strategy<KEY>* set_load_strategy(load_strategy<KEY>* strategy) {
        return get_instance().internal_set_load_strategy(strategy);
      }
      static const libraries& query_libraries() {
        return get_instance().internal_query_libraries();
      }
      static const keys& query_library(const std::string& library) {
        return get_instance().internal_query_library(library);
      }
      static const keys& query_keys() {
        return get_instance().internal_query_keys();
      }
      static void register_creator(const KEY& key, creator* c) {
        get_instance().internal_register_creator(key, c);
      }
      static void unregister_creator(const KEY& key, creator* c) {
        get_instance().internal_unregister_creator(key, c);
      }
      //==================================================
      // Instance handling
      //==================================================
    private:
      static KAPI factory<KEY,BASE>& get_instance();
# ifdef REGIME_FACTORY_PARAMS
      virtual BASE* internal_create(const KEY&, const parameters&) = 0;
# endif
      virtual BASE* internal_create(const KEY&) = 0;
      virtual void internal_clean() = 0;
      virtual void internal_clean(const KEY&) = 0;
      virtual void internal_load_key(const KEY&) = 0;
      virtual const std::string& internal_get_current_library() const = 0;
      virtual void internal_load_library(const std::string& library, bool is_auto) = 0;
      virtual void internal_unload_library(const std::string& library) = 0;
      virtual void internal_unload_key(const KEY&) = 0;
      virtual load_strategy<KEY>* internal_set_load_strategy(load_strategy<KEY>*) = 0;
      virtual const libraries& internal_query_libraries() const = 0;
      virtual const keys& internal_query_library(const std::string& library) const = 0;
      virtual const keys& internal_query_keys() const = 0;
      virtual void internal_register_creator(const KEY&, creator*) = 0;
      virtual void internal_unregister_creator(const KEY&, creator*) = 0;
    };

  //==================================================
  // IMPLEMENTATION OF: creator
  //==================================================
  template <class KEY, class BASE, class PRODUCT>
    creator<KEY,BASE,PRODUCT>::creator(const KEY& key)
    : creator_base<BASE>(factory<KEY,BASE>::get_current_library())
    , m_key(key)
    {
      factory<KEY,BASE>::register_creator(key, this);
    }

  template <class KEY, class BASE, class PRODUCT>
    creator<KEY,BASE,PRODUCT>::~creator()
    {
      factory<KEY,BASE>::unregister_creator(get_key(), this);
    }

  //==================================================
  // IMPLEMENTATION OF: singleton_creator
  //==================================================
  template <class KEY, class BASE, class PRODUCT>
    singleton_creator<KEY,BASE,PRODUCT>::singleton_creator(const KEY& key)
    : creator_base<BASE>(factory<KEY,BASE>::get_current_library())
    , m_key(key)
    , m_product(0)
    {
      factory<KEY,BASE>::register_creator(key, this);
    }

  template <class KEY, class BASE, class PRODUCT>
    singleton_creator<KEY,BASE,PRODUCT>::~singleton_creator()
    {
      // If I am destroyed, it means my library is unloaded.
      // In this case, I want to make sure my product does not exist
      if (get_library() != "")
        assert(m_product == 0);
      factory<KEY,BASE>::unregister_creator(get_key(), this);
    }
}
